@@ -1,6 +1,6 @@
 
 /*
- * $Id: authenticate.c,v 1.36.2.8 2003/02/04 23:31:42 robertc Exp $
+ * $Id: authenticate.c,v 1.48 2006/07/30 23:27:03 hno Exp $
  *
  * DEBUG: section 29    Authenticator
  * AUTHOR: Duane Wessels
@@ -42,8 +42,7 @@
 
 CBDATA_TYPE(auth_user_ip_t);
 
-static void
-     authenticateDecodeAuth(const char *proxy_auth, auth_user_request_t * auth_user_request);
+static void authenticateDecodeAuth(const char *proxy_auth, auth_user_request_t * auth_user_request);
 static auth_acl_t authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type headertype, request_t * request, ConnStateData * conn, struct in_addr src_addr);
 
 /*
@@ -83,7 +82,7 @@ authenticateAuthSchemeId(const char *typestr)
     return -1;
 }
 
-void
+static void
 authenticateDecodeAuth(const char *proxy_auth, auth_user_request_t * auth_user_request)
 {
     int i = 0;
@@ -94,7 +93,11 @@ authenticateDecodeAuth(const char *proxy_auth, auth_user_request_t * auth_user_r
 	/* we're configured to use this scheme - but is it active ? */
 	if ((i = authenticateAuthSchemeId(proxy_auth)) != -1) {
 	    authscheme_list[i].decodeauth(auth_user_request, proxy_auth);
-	    auth_user_request->auth_user->auth_module = i + 1;
+	    if (auth_user_request->auth_user) {
+		auth_user_request->auth_user->auth_module = i + 1;
+	    } else {
+		debug(29, 1) ("authenticateDecodeAuth: Invalid proxy-auth header, '%s'\n", proxy_auth);
+	    }
 	    return;
 	}
     }
@@ -349,7 +352,7 @@ authenticateAuthUserRequestClearIp(auth_user_request_t * auth_user_request)
 	authenticateAuthUserClearIp(auth_user_request->auth_user);
 }
 
-size_t
+int
 authenticateAuthUserRequestIPCount(auth_user_request_t * auth_user_request)
 {
     assert(auth_user_request);
@@ -381,6 +384,8 @@ authenticateUserAuthenticated(auth_user_request_t * auth_user_request)
 {
     if (!authenticateValidateUser(auth_user_request))
 	return 0;
+    if (auth_user_request->lastReply == AUTH_AUTHENTICATED)
+	return 1;
     if (auth_user_request->auth_user->auth_module > 0)
 	return authscheme_list[auth_user_request->auth_user->auth_module - 1].authenticated(auth_user_request);
     else
@@ -402,10 +407,12 @@ authenticateAuthenticateUser(auth_user_request_t * auth_user_request, request_t
 }
 
 static auth_user_request_t *
-authTryGetUser(auth_user_request_t ** auth_user_request, ConnStateData * conn)
+authTryGetUser(auth_user_request_t ** auth_user_request, ConnStateData * conn, request_t * request)
 {
     if (*auth_user_request)
 	return *auth_user_request;
+    else if (request && request->auth_user_request)
+	return request->auth_user_request;
     else if (conn)
 	return conn->auth_user_request;
     else
@@ -445,12 +452,12 @@ authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type
      * authenticated connection so we test for an authenticated
      * connection when we recieve no authentication header.
      */
-    if (((proxy_auth == NULL) && (!authenticateUserAuthenticated(authTryGetUser(auth_user_request, conn))))
+    if (((proxy_auth == NULL) && (!authenticateUserAuthenticated(authTryGetUser(auth_user_request, conn, request))))
 	|| (conn && conn->auth_type == AUTH_BROKEN)) {
 	/* no header or authentication failed/got corrupted - restart */
 	if (conn)
 	    conn->auth_type = AUTH_UNKNOWN;
-	debug(28, 4) ("authenticateAuthenticate: broken auth or no proxy_auth header. Requesting auth header.\n");
+	debug(29, 4) ("authenticateAuthenticate: broken auth or no proxy_auth header. Requesting auth header.\n");
 	/* something wrong with the AUTH credentials. Force a new attempt */
 	if (conn && conn->auth_user_request) {
 	    authenticateAuthUserRequestUnlock(conn->auth_user_request);
@@ -459,10 +466,11 @@ authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type
 	if (*auth_user_request) {
 	    /* unlock the ACL lock */
 	    authenticateAuthUserRequestUnlock(*auth_user_request);
-	    auth_user_request = NULL;
+	    *auth_user_request = NULL;
 	}
 	return AUTH_ACL_CHALLENGE;
     }
+#if 0
     /* 
      * Is this an already authenticated connection with a new auth header?
      * No check for function required in the if: its compulsory for conn based 
@@ -471,7 +479,7 @@ authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type
     if (proxy_auth && conn && conn->auth_user_request &&
 	authenticateUserAuthenticated(conn->auth_user_request) &&
 	strcmp(proxy_auth, authscheme_list[conn->auth_user_request->auth_user->auth_module - 1].authConnLastHeader(conn->auth_user_request))) {
-	debug(28, 2) ("authenticateAuthenticate: DUPLICATE AUTH - authentication header on already authenticated connection!. AU %p, Current user '%s' proxy_auth %s\n", conn->auth_user_request, authenticateUserRequestUsername(conn->auth_user_request), proxy_auth);
+	debug(29, 2) ("authenticateAuthenticate: DUPLICATE AUTH - authentication header on already authenticated connection!. AU %p, Current user '%s' proxy_auth %s\n", conn->auth_user_request, authenticateUserRequestUsername(conn->auth_user_request), proxy_auth);
 	/* remove this request struct - the link is already authed and it can't be to 
 	 * reauth.
 	 */
@@ -487,16 +495,27 @@ authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type
 	/* Set the connection auth type */
 	conn->auth_type = AUTH_UNKNOWN;
     }
+#endif
     /* we have a proxy auth header and as far as we know this connection has
      * not had bungled connection oriented authentication happen on it. */
-    debug(28, 9) ("authenticateAuthenticate: header %s.\n", proxy_auth);
+    debug(29, 9) ("authenticateAuthenticate: header %s.\n", proxy_auth ? proxy_auth : NULL);
     if (*auth_user_request == NULL) {
-	debug(28, 9) ("authenticateAuthenticate: This is a new checklist test on FD:%d\n",
+	debug(29, 9) ("authenticateAuthenticate: This is a new checklist test on FD:%d\n",
 	    conn ? conn->fd : -1);
+	if (proxy_auth && !request->auth_user_request && conn && conn->auth_user_request) {
+	    int id = authenticateAuthSchemeId(proxy_auth) + 1;
+	    if (!conn->auth_user_request->auth_user || conn->auth_user_request->auth_user->auth_module != id) {
+		debug(29, 1) ("authenticateAuthenticate: Unexpected change of authentication scheme from '%s' to '%s' (client %s)\n",
+		    authscheme_list[conn->auth_user_request->auth_user->auth_module - 1].typestr, proxy_auth, inet_ntoa(src_addr));
+		authenticateAuthUserRequestUnlock(conn->auth_user_request);
+		conn->auth_user_request = NULL;
+		conn->auth_type = AUTH_UNKNOWN;
+	    }
+	}
 	if ((!request->auth_user_request)
 	    && (!conn || conn->auth_type == AUTH_UNKNOWN)) {
 	    /* beginning of a new request check */
-	    debug(28, 4) ("authenticateAuthenticate: no connection authentication type\n");
+	    debug(29, 4) ("authenticateAuthenticate: no connection authentication type\n");
 	    if (!authenticateValidateUser(*auth_user_request =
 		    authenticateGetAuthUser(proxy_auth))) {
 		/* the decode might have left a username for logging, or a message to
@@ -524,7 +543,7 @@ authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type
 		authenticateAuthUserRequestLock(*auth_user_request);
 	    } else {
 		/* failed connection based authentication */
-		debug(28, 4) ("authenticateAuthenticate: Auth user request %p conn-auth user request %p conn type %d authentication failed.\n",
+		debug(29, 4) ("authenticateAuthenticate: Auth user request %p conn-auth user request %p conn type %d authentication failed.\n",
 		    *auth_user_request, conn->auth_user_request, conn->auth_type);
 		authenticateAuthUserRequestUnlock(*auth_user_request);
 		*auth_user_request = NULL;
@@ -538,6 +557,12 @@ authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type
 	    conn, headertype);
 	switch (authenticateDirection(*auth_user_request)) {
 	case 1:
+	    if (!request->auth_user_request) {
+		/* lock the user for the request structure link */
+		authenticateAuthUserRequestLock(*auth_user_request);
+		request->auth_user_request = *auth_user_request;
+	    }
+	    /* fallthrough to -2 */
 	case -2:
 	    /* this ACL check is finished. Unlock. */
 	    authenticateAuthUserRequestUnlock(*auth_user_request);
@@ -575,6 +600,7 @@ authenticateAuthenticate(auth_user_request_t ** auth_user_request, http_hdr_type
     }
     /* Unlock the request - we've authenticated it */
     authenticateAuthUserRequestUnlock(*auth_user_request);
+    *auth_user_request = NULL;
     return AUTH_AUTHENTICATED;
 }
 
@@ -582,20 +608,27 @@ auth_acl_t
 authenticateTryToAuthenticateAndSetAuthUser(auth_user_request_t ** auth_user_request, http_hdr_type headertype, request_t * request, ConnStateData * conn, struct in_addr src_addr)
 {
     /* If we have already been called, return the cached value */
-    auth_user_request_t *t = authTryGetUser(auth_user_request, conn);
+    auth_user_request_t *t = authTryGetUser(auth_user_request, conn, request);
     auth_acl_t result;
     if (t && t->lastReply != AUTH_ACL_CANNOT_AUTHENTICATE
 	&& t->lastReply != AUTH_ACL_HELPER) {
-	if (!*auth_user_request)
+	if (!*auth_user_request) {
 	    *auth_user_request = t;
+	    authenticateAuthUserRequestLock(*auth_user_request);
+	}
+	if (!request->auth_user_request) {
+	    request->auth_user_request = t;
+	    authenticateAuthUserRequestLock(request->auth_user_request);
+	}
 	return t->lastReply;
     }
     /* ok, call the actual authenticator routine. */
     result = authenticateAuthenticate(auth_user_request, headertype, request, conn, src_addr);
-    t = authTryGetUser(auth_user_request, conn);
+    t = authTryGetUser(auth_user_request, conn, request);
     if (t && result != AUTH_ACL_CANNOT_AUTHENTICATE &&
-	result != AUTH_ACL_HELPER)
+	result != AUTH_ACL_HELPER) {
 	t->lastReply = result;
+    }
     return result;
 }
 
@@ -667,6 +700,19 @@ authenticateSchemeInit(void)
     authSchemeSetup();
 }
 
+void
+authenticateConfigure(authConfig * config)
+{
+    int i;
+    authScheme *scheme;
+    for (i = 0; i < config->n_configured; i++) {
+	scheme = config->schemes + i;
+	if (authscheme_list[scheme->Id].checkconfig && authscheme_list[scheme->Id].configured()) {
+	    authscheme_list[scheme->Id].checkconfig(scheme);
+	}
+    }
+}
+
 void
 authenticateInit(authConfig * config)
 {
@@ -722,7 +768,7 @@ authenticateFixHeader(HttpReply * rep, auth_user_request_t * auth_user_request,
 	    || (rep->sline.status == HTTP_UNAUTHORIZED)) && internal)
 	/* this is a authenticate-needed response */
     {
-	if ((auth_user_request != NULL) && (auth_user_request->auth_user->auth_module > 0) & !authenticateUserAuthenticated(auth_user_request))
+	if ((auth_user_request != NULL) && (auth_user_request->auth_user->auth_module > 0) && authenticateDirection(auth_user_request) == 1)
 	    authscheme_list[auth_user_request->auth_user->auth_module - 1].authFixHeader(auth_user_request, rep, type, request);
 	else {
 	    int i;
@@ -730,11 +776,16 @@ authenticateFixHeader(HttpReply * rep, auth_user_request_t * auth_user_request,
 	    /* call each configured & running authscheme */
 	    for (i = 0; i < Config.authConfig.n_configured; i++) {
 		scheme = Config.authConfig.schemes + i;
-		if (authscheme_list[scheme->Id].Active())
-		    authscheme_list[scheme->Id].authFixHeader(NULL, rep, type,
-			request);
-		else
+		if (authscheme_list[scheme->Id].Active()) {
+		    if (auth_user_request && scheme->Id == auth_user_request->auth_user->auth_module - 1)
+			authscheme_list[scheme->Id].authFixHeader(
+			    auth_user_request, rep, type, request);
+		    else
+			authscheme_list[scheme->Id].authFixHeader(
+			    NULL, rep, type, request);
+		} else {
 		    debug(29, 4) ("authenticateFixHeader: Configured scheme %s not Active\n", scheme->typestr);
+		}
 	    }
 	}
     }
@@ -775,7 +826,7 @@ authenticateAuthUserUnlock(auth_user_t * auth_user)
     if (auth_user->references > 0) {
 	auth_user->references--;
     } else {
-	debug(29, 1) ("Attempt to lower Auth User %p refcount below 0!\n", auth_user);
+	fatalf("Attempt to lower Auth User %p refcount below 0!\n", auth_user);
     }
     debug(29, 9) ("authenticateAuthUserUnlock auth_user '%p' now at '%ld'.\n", auth_user, (long int) auth_user->references);
     if (auth_user->references == 0)
