@@ -1,6 +1,6 @@
 
 /*
- * $Id: store_dir_diskd.c,v 1.58.2.7 2003/01/09 03:38:44 hno Exp $
+ * $Id: store_dir_diskd.c,v 1.81 2006/07/30 23:47:14 hno Exp $
  *
  * DEBUG: section 47    Store Directory Routines
  * AUTHOR: Duane Wessels
@@ -86,7 +86,7 @@ static EVH storeDiskdDirRebuildFromSwapLog;
 static int storeDiskdDirGetNextFile(RebuildState *, sfileno *, int *size);
 static StoreEntry *storeDiskdDirAddDiskRestore(SwapDir * SD, const cache_key * key,
     sfileno file_number,
-    size_t swap_file_sz,
+    squid_file_sz swap_file_sz,
     time_t expires,
     time_t timestamp,
     time_t lastref,
@@ -99,6 +99,7 @@ static void storeDiskdDirCloseTmpSwapLog(SwapDir * sd);
 static FILE *storeDiskdDirOpenTmpSwapLog(SwapDir *, int *, int *);
 static STLOGOPEN storeDiskdDirOpenSwapLog;
 static STINIT storeDiskdDirInit;
+static STCHECKCONFIG storeDiskdCheckConfig;
 static STFREE storeDiskdDirFree;
 static STLOGCLEANSTART storeDiskdDirWriteCleanStart;
 static STLOGCLEANNEXTENTRY storeDiskdDirCleanLogNextEntry;
@@ -110,6 +111,7 @@ static STNEWFS storeDiskdDirNewfs;
 static STDUMP storeDiskdDirDump;
 static STMAINTAINFS storeDiskdDirMaintain;
 static STCHECKOBJ storeDiskdDirCheckObj;
+static STCHECKLOADAV storeDiskdDirCheckLoadAv;
 static STREFOBJ storeDiskdDirRefObj;
 static STUNREFOBJ storeDiskdDirUnrefObj;
 static QS rev_int_sort;
@@ -326,7 +328,7 @@ storeDiskdDirOpenSwapLog(SwapDir * sd)
     char *path;
     int fd;
     path = storeDiskdDirSwapLogFile(sd, NULL);
-    fd = file_open(path, O_WRONLY | O_CREAT);
+    fd = file_open(path, O_WRONLY | O_CREAT | O_BINARY);
     if (fd < 0) {
 	debug(50, 1) ("%s: %s\n", path, xstrerror());
 	fatal("storeDiskdDirOpenSwapLog: Failed to open swap log.");
@@ -355,13 +357,26 @@ storeDiskdDirCloseSwapLog(SwapDir * sd)
 	safe_free(diskd_dir_index);
 }
 
+static void
+storeDiskdCheckConfig(SwapDir * sd)
+{
+    requirePathnameExists("diskd_program", Config.Program.diskd);
+    if (!opt_create_swap_dirs)
+	requirePathnameExists("cache_dir", sd->path);
+}
+
+static void
+diskdExited(int fd, void *unused)
+{
+    fatal("diskd exited unexpectedly");
+}
+
 static void
 storeDiskdDirInit(SwapDir * sd)
 {
     static int started_clean_event = 0;
     int x;
     int i;
-    int rfd;
     int ikey;
     const char *args[5];
     char skey1[32];
@@ -385,8 +400,9 @@ storeDiskdDirInit(SwapDir * sd)
 	debug(50, 0) ("storeDiskdInit: msgget: %s\n", xstrerror());
 	fatal("msgget failed");
     }
+    diskdinfo->shm.nbufs = diskdinfo->magic2 * 1.3;
     diskdinfo->shm.id = shmget((key_t) (ikey + 2),
-	SHMBUFS * SHMBUF_BLKSZ, 0600 | IPC_CREAT);
+	diskdinfo->shm.nbufs * SHMBUF_BLKSZ, 0600 | IPC_CREAT);
     if (diskdinfo->shm.id < 0) {
 	debug(50, 0) ("storeDiskdInit: shmget: %s\n", xstrerror());
 	fatal("shmget failed");
@@ -396,9 +412,9 @@ storeDiskdDirInit(SwapDir * sd)
 	debug(50, 0) ("storeDiskdInit: shmat: %s\n", xstrerror());
 	fatal("shmat failed");
     }
-    diskdinfo->shm.inuse_map = xcalloc((SHMBUFS + 7) / 8, 1);
-    diskd_stats.shmbuf_count += SHMBUFS;
-    for (i = 0; i < SHMBUFS; i++) {
+    diskdinfo->shm.inuse_map = xcalloc((diskdinfo->shm.nbufs + 7) / 8, 1);
+    diskd_stats.shmbuf_count += diskdinfo->shm.nbufs;
+    for (i = 0; i < diskdinfo->shm.nbufs; i++) {
 	CBIT_SET(diskdinfo->shm.inuse_map, i);
 	storeDiskdShmPut(sd, i * SHMBUF_BLKSZ);
     }
@@ -410,24 +426,17 @@ storeDiskdDirInit(SwapDir * sd)
     args[2] = skey2;
     args[3] = skey3;
     args[4] = NULL;
-#if HAVE_POLL && defined(_SQUID_OSF_)
-    /* pipes and poll() don't get along on DUNIX -DW */
-    x = ipcCreate(IPC_TCP_SOCKET,
-#else
-    x = ipcCreate(IPC_FIFO,
-#endif
+    x = ipcCreate(IPC_STREAM,
 	Config.Program.diskd,
 	args,
 	"diskd",
-	&rfd,
+	&diskdinfo->rfd,
 	&diskdinfo->wfd);
     if (x < 0)
 	fatalf("execl: %s", Config.Program.diskd);
-    if (rfd != diskdinfo->wfd)
-	comm_close(rfd);
-    fd_note(diskdinfo->wfd, "squid -> diskd");
-    commSetTimeout(diskdinfo->wfd, -1, NULL, NULL);
-    commSetNonBlocking(diskdinfo->wfd);
+    fd_note(diskdinfo->rfd, "diskd -> squid health monitor");
+    fd_note(diskdinfo->wfd, "squid -> diskd health monitor");
+    commSetSelect(diskdinfo->rfd, COMM_SELECT_READ, diskdExited, NULL, 0);
     storeDiskdDirInitBitmap(sd);
     if (storeDiskdDirVerifyCacheDirs(sd) < 0)
 	fatal(errmsg);
@@ -452,18 +461,18 @@ storeDiskdStats(StoreEntry * sentry)
     storeAppendPrintf(sentry, "open_fail_queue_len: %d\n", diskd_stats.open_fail_queue_len);
     storeAppendPrintf(sentry, "block_queue_len: %d\n", diskd_stats.block_queue_len);
     diskd_stats.max_away = diskd_stats.max_shmuse = 0;
-    storeAppendPrintf(sentry, "\n             OPS SUCCESS    FAIL\n");
-    storeAppendPrintf(sentry, "%7s %7d %7d %7d\n",
+    storeAppendPrintf(sentry, "\n              OPS   SUCCESS    FAIL\n");
+    storeAppendPrintf(sentry, "%7s %9d %9d %7d\n",
 	"open", diskd_stats.open.ops, diskd_stats.open.success, diskd_stats.open.fail);
-    storeAppendPrintf(sentry, "%7s %7d %7d %7d\n",
+    storeAppendPrintf(sentry, "%7s %9d %9d %7d\n",
 	"create", diskd_stats.create.ops, diskd_stats.create.success, diskd_stats.create.fail);
-    storeAppendPrintf(sentry, "%7s %7d %7d %7d\n",
+    storeAppendPrintf(sentry, "%7s %9d %9d %7d\n",
 	"close", diskd_stats.close.ops, diskd_stats.close.success, diskd_stats.close.fail);
-    storeAppendPrintf(sentry, "%7s %7d %7d %7d\n",
+    storeAppendPrintf(sentry, "%7s %9d %9d %7d\n",
 	"unlink", diskd_stats.unlink.ops, diskd_stats.unlink.success, diskd_stats.unlink.fail);
-    storeAppendPrintf(sentry, "%7s %7d %7d %7d\n",
+    storeAppendPrintf(sentry, "%7s %9d %9d %7d\n",
 	"read", diskd_stats.read.ops, diskd_stats.read.success, diskd_stats.read.fail);
-    storeAppendPrintf(sentry, "%7s %7d %7d %7d\n",
+    storeAppendPrintf(sentry, "%7s %9d %9d %7d\n",
 	"write", diskd_stats.write.ops, diskd_stats.write.success, diskd_stats.write.fail);
 }
 
@@ -528,13 +537,31 @@ storeDiskdDirCallback(SwapDir * SD)
 	storeDiskdHandle(&M);
 	retval = 1;		/* Return that we've actually done some work */
 	if (M.shm_offset > -1)
-	    storeDiskdShmPut(SD, (off_t) M.shm_offset);
+	    storeDiskdShmPut(SD, M.shm_offset);
     }
     return retval;
 }
 
 
 
+static void
+storeDiskdDirRebuildComplete(RebuildState * rb)
+{
+    if (rb->log) {
+	debug(47, 1) ("Done reading %s swaplog (%d entries)\n",
+	    rb->sd->path, rb->n_read);
+	fclose(rb->log);
+	rb->log = NULL;
+    } else {
+	debug(47, 1) ("Done scanning %s (%d entries)\n",
+	    rb->sd->path, rb->counts.scancount);
+    }
+    store_dirs_rebuilding--;
+    storeDiskdDirCloseTmpSwapLog(rb->sd);
+    storeRebuildComplete(&rb->counts);
+    cbdataFree(rb);
+}
+
 static void
 storeDiskdDirRebuildFromDirectory(void *data)
 {
@@ -558,12 +585,7 @@ storeDiskdDirRebuildFromDirectory(void *data)
 	assert(fd == -1);
 	fd = storeDiskdDirGetNextFile(rb, &filn, &size);
 	if (fd == -2) {
-	    debug(20, 1) ("Done scanning %s swaplog (%d entries)\n",
-		rb->sd->path, rb->n_read);
-	    store_dirs_rebuilding--;
-	    storeDiskdDirCloseTmpSwapLog(rb->sd);
-	    storeRebuildComplete(&rb->counts);
-	    cbdataFree(rb);
+	    storeDiskdDirRebuildComplete(rb);
 	    return;
 	} else if (fd < 0) {
 	    continue;
@@ -615,10 +637,39 @@ storeDiskdDirRebuildFromDirectory(void *data)
 		assert(t->length == MD5_DIGEST_CHARS);
 		xmemcpy(key, t->value, MD5_DIGEST_CHARS);
 		break;
+#if SIZEOF_SQUID_FILE_SZ == SIZEOF_SIZE_T
 	    case STORE_META_STD:
 		assert(t->length == STORE_HDR_METASIZE);
 		xmemcpy(&tmpe.timestamp, t->value, STORE_HDR_METASIZE);
 		break;
+#else
+	    case STORE_META_STD_LFS:
+		assert(t->length == STORE_HDR_METASIZE);
+		xmemcpy(&tmpe.timestamp, t->value, STORE_HDR_METASIZE);
+		break;
+	    case STORE_META_STD:
+		assert(t->length == STORE_HDR_METASIZE_OLD);
+		{
+		    struct {
+			time_t timestamp;
+			time_t lastref;
+			time_t expires;
+			time_t lastmod;
+			size_t swap_file_sz;
+			u_short refcount;
+			u_short flags;
+		    }     *tmp = t->value;
+		    assert(sizeof(*tmp) == STORE_HDR_METASIZE_OLD);
+		    tmpe.timestamp = tmp->timestamp;
+		    tmpe.lastref = tmp->lastref;
+		    tmpe.expires = tmp->expires;
+		    tmpe.lastmod = tmp->lastmod;
+		    tmpe.swap_file_sz = tmp->swap_file_sz;
+		    tmpe.refcount = tmp->refcount;
+		    tmpe.flags = tmp->flags;
+		}
+		break;
+#endif
 	    default:
 		break;
 	    }
@@ -692,21 +743,10 @@ storeDiskdDirRebuildFromSwapLog(void *data)
     /* load a number of objects per invocation */
     for (count = 0; count < rb->speed; count++) {
 	if (fread(&s, ss, 1, rb->log) != 1) {
-	    debug(20, 1) ("Done reading %s swaplog (%d entries)\n",
-		rb->sd->path, rb->n_read);
-	    fclose(rb->log);
-	    rb->log = NULL;
-	    store_dirs_rebuilding--;
-	    storeDiskdDirCloseTmpSwapLog(rb->sd);
-	    storeRebuildComplete(&rb->counts);
-	    cbdataFree(rb);
+	    storeDiskdDirRebuildComplete(rb);
 	    return;
 	}
 	rb->n_read++;
-	if (s.op <= SWAP_LOG_NOP)
-	    continue;
-	if (s.op >= SWAP_LOG_MAX)
-	    continue;
 	/*
 	 * BC: during 2.4 development, we changed the way swap file
 	 * numbers are assigned and stored.  The high 16 bits used
@@ -757,15 +797,7 @@ storeDiskdDirRebuildFromSwapLog(void *data)
 		 * because adding to store_swap_size happens in
 		 * the cleanup procedure.
 		 */
-		storeExpireNow(e);
-		storeReleaseRequest(e);
-		if (e->swap_filen > -1) {
-		    storeDiskdDirReplRemove(e);
-		    storeDiskdDirMapBitReset(SD, e->swap_filen);
-		    e->swap_filen = -1;
-		    e->swap_dirn = -1;
-		}
-		storeRelease(e);
+		storeRecycle(e);
 		rb->counts.objcount--;
 		rb->counts.cancelcount++;
 	    }
@@ -845,16 +877,7 @@ storeDiskdDirRebuildFromSwapLog(void *data)
 	} else if (e) {
 	    /* key already exists, this swapfile not being used */
 	    /* junk old, load new */
-	    storeExpireNow(e);
-	    storeReleaseRequest(e);
-	    if (e->swap_filen > -1) {
-		storeDiskdDirReplRemove(e);
-		/* Make sure we don't actually unlink the file */
-		storeDiskdDirMapBitReset(SD, e->swap_filen);
-		e->swap_filen = -1;
-		e->swap_dirn = -1;
-	    }
-	    storeRelease(e);
+	    storeRecycle(e);
 	    rb->counts.dupcount++;
 	} else {
 	    /* URL doesnt exist, swapfile not in use */
@@ -878,6 +901,224 @@ storeDiskdDirRebuildFromSwapLog(void *data)
     eventAdd("storeRebuild", storeDiskdDirRebuildFromSwapLog, rb, 0.0, 1);
 }
 
+#if SIZEOF_SQUID_FILE_SZ != SIZEOF_SIZE_T
+/* This is an exact copy of the above, but using storeSwapLogDataOld entry type */
+static void
+storeDiskdDirRebuildFromSwapLogOld(void *data)
+{
+    RebuildState *rb = data;
+    SwapDir *SD = rb->sd;
+    StoreEntry *e = NULL;
+    storeSwapLogDataOld s;
+    size_t ss = sizeof(storeSwapLogDataOld);
+    int count;
+    int used;			/* is swapfile already in use? */
+    int disk_entry_newer;	/* is the log entry newer than current entry? */
+    double x;
+    assert(rb != NULL);
+    /* load a number of objects per invocation */
+    for (count = 0; count < rb->speed; count++) {
+	if (fread(&s, ss, 1, rb->log) != 1) {
+	    storeDiskdDirRebuildComplete(rb);
+	    return;
+	}
+	rb->n_read++;
+	/*
+	 * BC: during 2.4 development, we changed the way swap file
+	 * numbers are assigned and stored.  The high 16 bits used
+	 * to encode the SD index number.  There used to be a call
+	 * to storeDirProperFileno here that re-assigned the index
+	 * bits.  Now, for backwards compatibility, we just need
+	 * to mask it off.
+	 */
+	s.swap_filen &= 0x00FFFFFF;
+	debug(20, 3) ("storeDiskdDirRebuildFromSwapLog: %s %s %08X\n",
+	    swap_log_op_str[(int) s.op],
+	    storeKeyText(s.key),
+	    s.swap_filen);
+	if (s.op == SWAP_LOG_ADD) {
+	    /*
+	     * Here we have some special checks for large files.
+	     * I've been seeing a system crash followed by a reboot
+	     * that seems to corrupt the swap log.  Squid believes
+	     * that the disk holds some really large files.  It
+	     * complains about using being over the high water mark
+	     * and proceeds to delete files as fast as it can.  To
+	     * prevent that, we call stat() on sufficiently large
+	     * files (>128KB) and reject those that are missing or
+	     * have the wrong size.
+	     */
+	    struct stat sb;
+	    char *p = storeDiskdDirFullPath(SD, s.swap_filen, NULL);
+	    if (s.swap_file_sz < (1 << 17)) {
+		(void) 0;
+	    } else if (stat(p, &sb) < 0) {
+		debug(47, 2) ("its missing!: %s\n", p);
+		continue;
+	    } else if (sb.st_size != s.swap_file_sz) {
+		debug(47, 2) ("size mismatch!: stat=%d, log=%d\n",
+		    (int) sb.st_size, (int) s.swap_file_sz);
+		continue;
+	    } else {
+		debug(47, 2) ("big file (%d bytes) checks out\n",
+		    (int) s.swap_file_sz);
+	    }
+	} else if (s.op == SWAP_LOG_DEL) {
+	    /* Delete unless we already have a newer copy */
+	    if ((e = storeGet(s.key)) != NULL && s.lastref > e->lastref) {
+		/*
+		 * Make sure we don't unlink the file, it might be
+		 * in use by a subsequent entry.  Also note that
+		 * we don't have to subtract from store_swap_size
+		 * because adding to store_swap_size happens in
+		 * the cleanup procedure.
+		 */
+		storeRecycle(e);
+		rb->counts.objcount--;
+		rb->counts.cancelcount++;
+	    }
+	    continue;
+	} else {
+	    x = log(++rb->counts.bad_log_op) / log(10.0);
+	    if (0.0 == x - (double) (int) x)
+		debug(20, 1) ("WARNING: %d invalid swap log entries found\n",
+		    rb->counts.bad_log_op);
+	    rb->counts.invalid++;
+	    continue;
+	}
+	if ((++rb->counts.scancount & 0xFFF) == 0) {
+	    struct stat sb;
+	    if (0 == fstat(fileno(rb->log), &sb))
+		storeRebuildProgress(SD->index,
+		    (int) sb.st_size / ss, rb->n_read);
+	}
+	if (!storeDiskdDirValidFileno(SD, s.swap_filen, 0)) {
+	    rb->counts.invalid++;
+	    continue;
+	}
+	if (EBIT_TEST(s.flags, KEY_PRIVATE)) {
+	    rb->counts.badflags++;
+	    continue;
+	}
+	e = storeGet(s.key);
+	used = storeDiskdDirMapBitTest(SD, s.swap_filen);
+	/* If this URL already exists in the cache, does the swap log
+	 * appear to have a newer entry?  Compare 'lastref' from the
+	 * swap log to e->lastref. */
+	disk_entry_newer = e ? (s.lastref > e->lastref ? 1 : 0) : 0;
+	if (used && !disk_entry_newer) {
+	    /* log entry is old, ignore it */
+	    rb->counts.clashcount++;
+	    continue;
+	} else if (used && e && e->swap_filen == s.swap_filen && e->swap_dirn == SD->index) {
+	    /* swapfile taken, same URL, newer, update meta */
+	    if (e->store_status == STORE_OK) {
+		e->lastref = s.timestamp;
+		e->timestamp = s.timestamp;
+		e->expires = s.expires;
+		e->lastmod = s.lastmod;
+		e->flags = s.flags;
+		e->refcount += s.refcount;
+		storeDiskdDirUnrefObj(SD, e);
+	    } else {
+		debug_trap("storeDiskdDirRebuildFromSwapLog: bad condition");
+		debug(20, 1) ("\tSee %s:%d\n", __FILE__, __LINE__);
+	    }
+	    continue;
+	} else if (used) {
+	    /* swapfile in use, not by this URL, log entry is newer */
+	    /* This is sorta bad: the log entry should NOT be newer at this
+	     * point.  If the log is dirty, the filesize check should have
+	     * caught this.  If the log is clean, there should never be a
+	     * newer entry. */
+	    debug(20, 1) ("WARNING: newer swaplog entry for dirno %d, fileno %08X\n",
+		SD->index, s.swap_filen);
+	    /* I'm tempted to remove the swapfile here just to be safe,
+	     * but there is a bad race condition in the NOVM version if
+	     * the swapfile has recently been opened for writing, but
+	     * not yet opened for reading.  Because we can't map
+	     * swapfiles back to StoreEntrys, we don't know the state
+	     * of the entry using that file.  */
+	    /* We'll assume the existing entry is valid, probably because
+	     * were in a slow rebuild and the the swap file number got taken
+	     * and the validation procedure hasn't run. */
+	    assert(rb->flags.need_to_validate);
+	    rb->counts.clashcount++;
+	    continue;
+	} else if (e && !disk_entry_newer) {
+	    /* key already exists, current entry is newer */
+	    /* keep old, ignore new */
+	    rb->counts.dupcount++;
+	    continue;
+	} else if (e) {
+	    /* key already exists, this swapfile not being used */
+	    /* junk old, load new */
+	    storeRecycle(e);
+	    rb->counts.dupcount++;
+	} else {
+	    /* URL doesnt exist, swapfile not in use */
+	    /* load new */
+	    (void) 0;
+	}
+	/* update store_swap_size */
+	rb->counts.objcount++;
+	e = storeDiskdDirAddDiskRestore(SD, s.key,
+	    s.swap_filen,
+	    s.swap_file_sz,
+	    s.expires,
+	    s.timestamp,
+	    s.lastref,
+	    s.lastmod,
+	    s.refcount,
+	    s.flags,
+	    (int) rb->flags.clean);
+	storeDirSwapLog(e, SWAP_LOG_ADD);
+    }
+    eventAdd("storeRebuild", storeDiskdDirRebuildFromSwapLogOld, rb, 0.0, 1);
+}
+
+#endif
+
+static void
+storeDiskdDirRebuildFromSwapLogCheckVersion(void *data)
+{
+    RebuildState *rb = data;
+    storeSwapLogHeader hdr;
+
+    if (fread(&hdr, sizeof(hdr), 1, rb->log) != 1) {
+	storeDiskdDirRebuildComplete(rb);
+	return;
+    }
+    if (hdr.op == SWAP_LOG_VERSION) {
+	if (fseek(rb->log, hdr.record_size, SEEK_SET) != 0) {
+	    storeDiskdDirRebuildComplete(rb);
+	    return;
+	}
+	if (hdr.version == 1 && hdr.record_size == sizeof(storeSwapLogData)) {
+	    eventAdd("storeRebuild", storeDiskdDirRebuildFromSwapLog, rb, 0.0, 1);
+	    return;
+	}
+#if SIZEOF_SQUID_FILE_SZ != SIZEOF_SIZE_T
+	if (hdr.version == 1 && hdr.record_size == sizeof(storeSwapLogDataOld)) {
+	    debug(47, 1) ("storeDiskdDirRebuildFromSwapLog: Found current version but without large file support. Upgrading\n");
+	    eventAdd("storeRebuild", storeDiskdDirRebuildFromSwapLogOld, rb, 0.0, 1);
+	    return;
+	}
+#endif
+	debug(47, 1) ("storeDiskdDirRebuildFromSwapLog: Unsupported swap.state version %d size %d\n",
+	    hdr.version, hdr.record_size);
+	storeDiskdDirRebuildComplete(rb);
+	return;
+    }
+    rewind(rb->log);
+    debug(47, 1) ("storeDiskdDirRebuildFromSwapLog: Old version detected. Upgrading\n");
+#if SIZEOF_SQUID_FILE_SZ == SIZEOF_SIZE_T
+    eventAdd("storeRebuild", storeDiskdDirRebuildFromSwapLog, rb, 0.0, 1);
+#else
+    eventAdd("storeRebuild", storeDiskdDirRebuildFromSwapLogOld, rb, 0.0, 1);
+#endif
+}
+
 static int
 storeDiskdDirGetNextFile(RebuildState * rb, sfileno * filn_p, int *size)
 {
@@ -942,7 +1183,7 @@ storeDiskdDirGetNextFile(RebuildState * rb, sfileno * filn_p, int *size)
 	    snprintf(rb->fullfilename, SQUID_MAXPATHLEN, "%s/%s",
 		rb->fullpath, rb->entry->d_name);
 	    debug(20, 3) ("storeDiskdDirGetNextFile: Opening %s\n", rb->fullfilename);
-	    fd = file_open(rb->fullfilename, O_RDONLY);
+	    fd = file_open(rb->fullfilename, O_RDONLY | O_BINARY);
 	    if (fd < 0)
 		debug(50, 1) ("storeDiskdDirGetNextFile: %s: %s\n", rb->fullfilename, xstrerror());
 	    else
@@ -970,7 +1211,7 @@ storeDiskdDirGetNextFile(RebuildState * rb, sfileno * filn_p, int *size)
 static StoreEntry *
 storeDiskdDirAddDiskRestore(SwapDir * SD, const cache_key * key,
     int file_number,
-    size_t swap_file_sz,
+    squid_file_sz swap_file_sz,
     time_t expires,
     time_t timestamp,
     time_t lastref,
@@ -1034,7 +1275,7 @@ storeDiskdDirRebuild(SwapDir * sd)
 	    fclose(fp);
 	func = storeDiskdDirRebuildFromDirectory;
     } else {
-	func = storeDiskdDirRebuildFromSwapLog;
+	func = storeDiskdDirRebuildFromSwapLogCheckVersion;
 	rb->log = fp;
 	rb->flags.clean = (unsigned int) clean;
     }
@@ -1054,16 +1295,10 @@ storeDiskdDirCloseTmpSwapLog(SwapDir * sd)
     char *new_path = xstrdup(storeDiskdDirSwapLogFile(sd, ".new"));
     int fd;
     file_close(diskdinfo->swaplog_fd);
-#ifdef _SQUID_OS2_
-    if (unlink(swaplog_path) < 0) {
-	debug(50, 0) ("%s: %s\n", swaplog_path, xstrerror());
-	fatal("storeDiskdDirCloseTmpSwapLog: unlink failed");
-    }
-#endif
     if (xrename(new_path, swaplog_path) < 0) {
 	fatal("storeDiskdDirCloseTmpSwapLog: rename failed");
     }
-    fd = file_open(swaplog_path, O_WRONLY | O_CREAT);
+    fd = file_open(swaplog_path, O_WRONLY | O_CREAT | O_BINARY);
     if (fd < 0) {
 	debug(50, 1) ("%s: %s\n", swaplog_path, xstrerror());
 	fatal("storeDiskdDirCloseTmpSwapLog: Failed to open swap log.");
@@ -1074,6 +1309,31 @@ storeDiskdDirCloseTmpSwapLog(SwapDir * sd)
     debug(47, 3) ("Cache Dir #%d log opened on FD %d\n", sd->index, fd);
 }
 
+static void
+storeSwapLogDataFree(void *s)
+{
+    memFree(s, MEM_SWAP_LOG_DATA);
+}
+
+static void
+storeDiskdWriteSwapLogheader(int fd)
+{
+    storeSwapLogHeader *hdr = memAllocate(MEM_SWAP_LOG_DATA);
+    hdr->op = SWAP_LOG_VERSION;
+    hdr->version = 1;
+    hdr->record_size = sizeof(storeSwapLogData);
+    /* The header size is a full log record to keep some level of backward
+     * compatibility even if the actual header is smaller
+     */
+    file_write(fd,
+	-1,
+	hdr,
+	sizeof(storeSwapLogData),
+	NULL,
+	NULL,
+	(FREE *) storeSwapLogDataFree);
+}
+
 static FILE *
 storeDiskdDirOpenTmpSwapLog(SwapDir * sd, int *clean_flag, int *zero_flag)
 {
@@ -1097,14 +1357,15 @@ storeDiskdDirOpenTmpSwapLog(SwapDir * sd, int *clean_flag, int *zero_flag)
     if (diskdinfo->swaplog_fd >= 0)
 	file_close(diskdinfo->swaplog_fd);
     /* open a write-only FD for the new log */
-    fd = file_open(new_path, O_WRONLY | O_CREAT | O_TRUNC);
+    fd = file_open(new_path, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY);
     if (fd < 0) {
 	debug(50, 1) ("%s: %s\n", new_path, xstrerror());
 	fatal("storeDirOpenTmpSwapLog: Failed to open swap log.");
     }
     diskdinfo->swaplog_fd = fd;
+    storeDiskdWriteSwapLogheader(fd);
     /* open a read-only stream of the old log */
-    fp = fopen(swaplog_path, "r");
+    fp = fopen(swaplog_path, "rb");
     if (fp == NULL) {
 	debug(50, 0) ("%s: %s\n", swaplog_path, xstrerror());
 	fatal("Failed to open swap log for reading");
@@ -1128,7 +1389,7 @@ struct _clean_state {
     char *new;
     char *cln;
     char *outbuf;
-    off_t outbuf_offset;
+    int outbuf_offset;
     int fd;
     RemovalPolicyWalker *walker;
 };
@@ -1155,7 +1416,7 @@ storeDiskdDirWriteCleanStart(SwapDir * sd)
     state->outbuf_offset = 0;
     state->walker = sd->repl->WalkInit(sd->repl);
     unlink(state->cln);
-    state->fd = file_open(state->new, O_WRONLY | O_CREAT | O_TRUNC);
+    state->fd = file_open(state->new, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY);
     if (state->fd < 0) {
 	xfree(state->new);
 	xfree(state->cur);
@@ -1165,6 +1426,7 @@ storeDiskdDirWriteCleanStart(SwapDir * sd)
     }
     debug(20, 3) ("storeDirWriteCleanLogs: opened %s, FD %d\n",
 	state->new, state->fd);
+    storeDiskdWriteSwapLogheader(state->fd);
 #if HAVE_FCHMOD
     if (stat(state->cur, &sb) == 0)
 	fchmod(state->fd, sb.st_mode);
@@ -1256,9 +1518,6 @@ storeDiskdDirWriteCleanDone(SwapDir * sd)
 #ifdef _SQUID_OS2_
 	file_close(state->fd);
 	state->fd = -1;
-	if (unlink(cur) < 0)
-	    debug(50, 0) ("storeDirWriteCleanLogs: unlinkd failed: %s, %s\n",
-		xstrerror(), cur);
 #endif
 	xrename(state->new, state->cur);
     }
@@ -1268,7 +1527,7 @@ storeDiskdDirWriteCleanDone(SwapDir * sd)
     else if (state->fd < 0)
 	(void) 0;
     else
-	file_close(file_open(state->cln, O_WRONLY | O_CREAT | O_TRUNC));
+	file_close(file_open(state->cln, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY));
     /* close */
     safe_free(state->cur);
     safe_free(state->new);
@@ -1281,12 +1540,6 @@ storeDiskdDirWriteCleanDone(SwapDir * sd)
     sd->log.clean.write = NULL;
 }
 
-static void
-storeSwapLogDataFree(void *s)
-{
-    memFree(s, MEM_SWAP_LOG_DATA);
-}
-
 static void
 storeDiskdDirSwapLog(const SwapDir * sd, const StoreEntry * e, int op)
 {
@@ -1547,9 +1800,18 @@ storeDiskdDirCheckObj(SwapDir * SD, const StoreEntry * e)
     diskdinfo_t *diskdinfo = SD->fsdata;
     /* Check the queue length */
     if (diskdinfo->away >= diskdinfo->magic1)
-	return -1;
+	return 0;
+    return 1;
+}
+
+int
+storeDiskdDirCheckLoadAv(SwapDir * SD, store_op_t op)
+{
+    diskdinfo_t *diskdinfo = SD->fsdata;
     /* Calculate the storedir load relative to magic2 on a scale of 0 .. 1000 */
     /* the parse function guarantees magic2 is positivie */
+    if (diskdinfo->away >= diskdinfo->magic1)
+	return -1;
     return diskdinfo->away * 1000 / diskdinfo->magic2;
 }
 
@@ -1636,12 +1898,12 @@ storeDiskdDirReplRemove(StoreEntry * e)
  */
 
 void *
-storeDiskdShmGet(SwapDir * sd, off_t * shm_offset)
+storeDiskdShmGet(SwapDir * sd, int *shm_offset)
 {
     char *buf = NULL;
     diskdinfo_t *diskdinfo = sd->fsdata;
     int i;
-    for (i = 0; i < SHMBUFS; i++) {
+    for (i = 0; i < diskdinfo->shm.nbufs; i++) {
 	if (CBIT_TEST(diskdinfo->shm.inuse_map, i))
 	    continue;
 	CBIT_SET(diskdinfo->shm.inuse_map, i);
@@ -1651,7 +1913,7 @@ storeDiskdShmGet(SwapDir * sd, off_t * shm_offset)
     }
     assert(buf);
     assert(buf >= diskdinfo->shm.buf);
-    assert(buf < diskdinfo->shm.buf + (SHMBUFS * SHMBUF_BLKSZ));
+    assert(buf < diskdinfo->shm.buf + (diskdinfo->shm.nbufs * SHMBUF_BLKSZ));
     diskd_stats.shmbuf_count++;
     if (diskd_stats.max_shmuse < diskd_stats.shmbuf_count)
 	diskd_stats.max_shmuse = diskd_stats.shmbuf_count;
@@ -1659,14 +1921,14 @@ storeDiskdShmGet(SwapDir * sd, off_t * shm_offset)
 }
 
 void
-storeDiskdShmPut(SwapDir * sd, off_t offset)
+storeDiskdShmPut(SwapDir * sd, int offset)
 {
     int i;
     diskdinfo_t *diskdinfo = sd->fsdata;
     assert(offset >= 0);
-    assert(offset < SHMBUFS * SHMBUF_BLKSZ);
+    assert(offset < diskdinfo->shm.nbufs * SHMBUF_BLKSZ);
     i = offset / SHMBUF_BLKSZ;
-    assert(i < SHMBUFS);
+    assert(i < diskdinfo->shm.nbufs);
     assert(CBIT_TEST(diskdinfo->shm.inuse_map, i));
     CBIT_CLR(diskdinfo->shm.inuse_map, i);
     diskd_stats.shmbuf_count--;
@@ -1681,10 +1943,17 @@ void
 storeDiskdDirStats(SwapDir * SD, StoreEntry * sentry)
 {
     diskdinfo_t *diskdinfo = SD->fsdata;
+#ifdef HAVE_STATVFS
+    fsblkcnt_t totl_kb = 0;
+    fsblkcnt_t free_kb = 0;
+    fsfilcnt_t totl_in = 0;
+    fsfilcnt_t free_in = 0;
+#else
     int totl_kb = 0;
     int free_kb = 0;
     int totl_in = 0;
     int free_in = 0;
+#endif
     int x;
     storeAppendPrintf(sentry, "First level subdirectories: %d\n", diskdinfo->l1);
     storeAppendPrintf(sentry, "Second level subdirectories: %d\n", diskdinfo->l2);
@@ -1697,6 +1966,16 @@ storeDiskdDirStats(SwapDir * SD, StoreEntry * sentry)
 	percent(diskdinfo->map->n_files_in_map, diskdinfo->map->max_n_files));
     x = storeDirGetUFSStats(SD->path, &totl_kb, &free_kb, &totl_in, &free_in);
     if (0 == x) {
+#ifdef HAVE_STATVFS
+	storeAppendPrintf(sentry, "Filesystem Space in use: %" PRIu64 "/%" PRIu64 " KB (%.0f%%)\n",
+	    (uint64_t) (totl_kb - free_kb),
+	    (uint64_t) totl_kb,
+	    dpercent(totl_kb - free_kb, totl_kb));
+	storeAppendPrintf(sentry, "Filesystem Inodes in use: %" PRIu64 "/%" PRIu64 " (%.0f%%)\n",
+	    (uint64_t) (totl_in - free_in),
+	    (uint64_t) totl_in,
+	    dpercent(totl_in - free_in, totl_in));
+#else
 	storeAppendPrintf(sentry, "Filesystem Space in use: %d/%d KB (%d%%)\n",
 	    totl_kb - free_kb,
 	    totl_kb,
@@ -1705,6 +1984,7 @@ storeDiskdDirStats(SwapDir * SD, StoreEntry * sentry)
 	    totl_in - free_in,
 	    totl_in,
 	    percent(totl_in - free_in, totl_in));
+#endif
     }
     storeAppendPrintf(sentry, "Flags:");
     if (SD->flags.selected)
@@ -1721,7 +2001,21 @@ storeDiskdDirParseQ1(SwapDir * sd, const char *name, const char *value, int reco
     diskdinfo_t *diskdinfo = sd->fsdata;
     int old_magic1 = diskdinfo->magic1;
     diskdinfo->magic1 = atoi(value);
-    if (reconfiguring && old_magic1 != diskdinfo->magic1)
+    if (!reconfiguring)
+	return;
+    if (old_magic1 < diskdinfo->magic1) {
+	/*
+	 * This is because shm.nbufs is computed at startup, when
+	 * we call shmget().  We can't increase the Q1/Q2 parameters
+	 * beyond their initial values because then we might have
+	 * more "Q2 messages" than shared memory chunks, and this
+	 * will cause an assertion in storeDiskdShmGet().
+	 */
+	debug(3, 1) ("WARNING: cannot increase cache_dir '%s' Q1 value while Squid is running.\n", sd->path);
+	diskdinfo->magic1 = old_magic1;
+	return;
+    }
+    if (old_magic1 != diskdinfo->magic1)
 	debug(3, 1) ("cache_dir '%s' new Q1 value '%d'\n",
 	    sd->path, diskdinfo->magic1);
 }
@@ -1739,7 +2033,15 @@ storeDiskdDirParseQ2(SwapDir * sd, const char *name, const char *value, int reco
     diskdinfo_t *diskdinfo = sd->fsdata;
     int old_magic2 = diskdinfo->magic2;
     diskdinfo->magic2 = atoi(value);
-    if (reconfiguring && old_magic2 != diskdinfo->magic2)
+    if (!reconfiguring)
+	return;
+    if (old_magic2 < diskdinfo->magic2) {
+	/* See comments in Q1 function above */
+	debug(3, 1) ("WARNING: cannot increase cache_dir '%s' Q2 value while Squid is running.\n", sd->path);
+	diskdinfo->magic2 = old_magic2;
+	return;
+    }
+    if (old_magic2 != diskdinfo->magic2)
 	debug(3, 1) ("cache_dir '%s' new Q2 value '%d'\n",
 	    sd->path, diskdinfo->magic2);
 }
@@ -1751,7 +2053,7 @@ storeDiskdDirDumpQ2(StoreEntry * e, const char *option, SwapDir * sd)
     storeAppendPrintf(e, " Q2=%d", diskdinfo->magic2);
 }
 
-struct cache_dir_option options[] =
+static struct cache_dir_option options[] =
 {
 #if NOT_YET
     {"L1", storeDiskdDirParseL1, storeDiskdDirDumpL1},
@@ -1913,6 +2215,7 @@ storeDiskdDirParse(SwapDir * sd, int index, char *path)
     diskdinfo->suggest = 0;
     diskdinfo->magic1 = 64;
     diskdinfo->magic2 = 72;
+    sd->checkconfig = storeDiskdCheckConfig;
     sd->init = storeDiskdDirInit;
     sd->newfs = storeDiskdDirNewfs;
     sd->dump = storeDiskdDirDump;
@@ -1921,6 +2224,7 @@ storeDiskdDirParse(SwapDir * sd, int index, char *path)
     sd->statfs = storeDiskdDirStats;
     sd->maintainfs = storeDiskdDirMaintain;
     sd->checkobj = storeDiskdDirCheckObj;
+    sd->checkload = storeDiskdDirCheckLoadAv;
     sd->refobj = storeDiskdDirRefObj;
     sd->unrefobj = storeDiskdDirUnrefObj;
     sd->callback = storeDiskdDirCallback;
@@ -1931,6 +2235,7 @@ storeDiskdDirParse(SwapDir * sd, int index, char *path)
     sd->obj.read = storeDiskdRead;
     sd->obj.write = storeDiskdWrite;
     sd->obj.unlink = storeDiskdUnlink;
+    sd->obj.recycle = storeDiskdRecycle;
     sd->log.open = storeDiskdDirOpenSwapLog;
     sd->log.close = storeDiskdDirCloseSwapLog;
     sd->log.write = storeDiskdDirSwapLog;
