@@ -1,6 +1,6 @@
 
 /*
- * $Id: tools.c,v 1.213.2.5 2003/02/08 02:31:10 robertc Exp $
+ * $Id: tools.c,v 1.244 2006/07/18 00:22:20 hno Exp $
  *
  * DEBUG: section 21    Misc Functions
  * AUTHOR: Harvest Derived
@@ -35,6 +35,19 @@
 
 #include "squid.h"
 
+#if LINUX_TPROXY
+#undef _POSIX_SOURCE
+/* Ugly glue to get around linux header madness colliding with glibc */
+#define _LINUX_TYPES_H
+#define _LINUX_FS_H
+typedef uint32_t __u32;
+#include <sys/capability.h>
+#endif
+
+#if HAVE_SYS_PRCTL_H
+#include <sys/prctl.h>
+#endif
+
 #define DEAD_MSG "\
 The Squid Cache (version %s) died.\n\
 \n\
@@ -106,11 +119,14 @@ mail_warranty(void)
     if ((fp = fopen(filename, "w")) == NULL)
 	return;
 #endif
-    fprintf(fp, "From: %s\n", appname);
+    if (Config.EmailFrom)
+	fprintf(fp, "From: %s\n", Config.EmailFrom);
+    else
+	fprintf(fp, "From: %s@%s\n", appname, uniqueHostname());
     fprintf(fp, "To: %s\n", Config.adminEmail);
     fprintf(fp, "Subject: %s\n", dead_msg());
     fclose(fp);
-    snprintf(command, 256, "mail %s < %s", Config.adminEmail, filename);
+    snprintf(command, 256, "%s %s < %s", Config.EmailProgram, Config.adminEmail, filename);
     system(command);		/* XXX should avoid system(3) */
     unlink(filename);
 }
@@ -132,18 +148,18 @@ dumpMallocStats(void)
 	return;
     mp = mallinfo();
     fprintf(debug_log, "Memory usage for %s via mallinfo():\n", appname);
-    fprintf(debug_log, "\ttotal space in arena:  %6d KB\n",
-	mp.arena >> 10);
-    fprintf(debug_log, "\tOrdinary blocks:       %6d KB %6d blks\n",
-	mp.uordblks >> 10, mp.ordblks);
-    fprintf(debug_log, "\tSmall blocks:          %6d KB %6d blks\n",
-	mp.usmblks >> 10, mp.smblks);
-    fprintf(debug_log, "\tHolding blocks:        %6d KB %6d blks\n",
-	mp.hblkhd >> 10, mp.hblks);
-    fprintf(debug_log, "\tFree Small blocks:     %6d KB\n",
-	mp.fsmblks >> 10);
-    fprintf(debug_log, "\tFree Ordinary blocks:  %6d KB\n",
-	mp.fordblks >> 10);
+    fprintf(debug_log, "\ttotal space in arena:  %6ld KB\n",
+	(long int) mp.arena >> 10);
+    fprintf(debug_log, "\tOrdinary blocks:       %6ld KB %6ld blks\n",
+	(long int) mp.uordblks >> 10, (long int) mp.ordblks);
+    fprintf(debug_log, "\tSmall blocks:          %6ld KB %6ld blks\n",
+	(long int) mp.usmblks >> 10, (long int) mp.smblks);
+    fprintf(debug_log, "\tHolding blocks:        %6ld KB %6ld blks\n",
+	(long int) mp.hblkhd >> 10, (long int) mp.hblks);
+    fprintf(debug_log, "\tFree Small blocks:     %6ld KB\n",
+	(long int) mp.fsmblks >> 10);
+    fprintf(debug_log, "\tFree Ordinary blocks:  %6ld KB\n",
+	(long int) mp.fordblks >> 10);
     t = mp.uordblks + mp.usmblks + mp.hblkhd;
     fprintf(debug_log, "\tTotal in use:          %6d KB %d%%\n",
 	t >> 10, percent(t, mp.arena));
@@ -206,6 +222,8 @@ rusage_maxrss(struct rusage *r)
     return r->ru_maxrss;
 #elif defined(_SQUID_OSF_)
     return r->ru_maxrss;
+#elif defined(_SQUID_AIX_)
+    return r->ru_maxrss;
 #elif defined(BSD4_4)
     return r->ru_maxrss;
 #elif defined(HAVE_GETPAGESIZE) && HAVE_GETPAGESIZE != 0
@@ -282,7 +300,7 @@ death(int sig)
 #endif
 #endif /* PRINT_STACK_TRACE */
 
-#if SA_RESETHAND == 0
+#if SA_RESETHAND == 0 && !defined(_SQUID_MSWIN_)
     signal(SIGSEGV, SIG_DFL);
     signal(SIGBUS, SIG_DFL);
     signal(sig, SIG_DFL);
@@ -339,8 +357,38 @@ fatal_common(const char *message)
     fprintf(debug_log, "Squid Cache (Version %s): Terminated abnormally.\n",
 	version_string);
     fflush(debug_log);
-    PrintRusage();
-    dumpMallocStats();
+#ifdef PRINT_STACK_TRACE
+#ifdef _SQUID_HPUX_
+    {
+	extern void U_STACK_TRACE(void);	/* link with -lcl */
+	fflush(debug_log);
+	dup2(fileno(debug_log), 2);
+	U_STACK_TRACE();
+    }
+#endif /* _SQUID_HPUX_ */
+#ifdef _SQUID_SOLARIS_
+    {				/* get ftp://opcom.sun.ca/pub/tars/opcom_stack.tar.gz and */
+	extern void opcom_stack_trace(void);	/* link with -lopcom_stack */
+	fflush(debug_log);
+	dup2(fileno(debug_log), fileno(stdout));
+	opcom_stack_trace();
+	fflush(stdout);
+    }
+#endif /* _SQUID_SOLARIS_ */
+#if HAVE_BACKTRACE_SYMBOLS_FD
+    {
+	static void *(callarray[8192]);
+	int n;
+	n = backtrace(callarray, 8192);
+	backtrace_symbols_fd(callarray, n, fileno(debug_log));
+    }
+#endif
+#endif /* PRINT_STACK_TRACE */
+
+    if (!shutting_down) {
+	PrintRusage();
+	dumpMallocStats();
+    }
 }
 
 /* fatal */
@@ -355,7 +403,7 @@ fatal(const char *message)
 	storeDirWriteCleanLogs(0);
     fatal_common(message);
     if (shutting_down)
-	exit(0);
+	exit(1);
     else
 	abort();
 }
@@ -415,6 +463,7 @@ debug_trap(const char *message)
 void
 sig_child(int sig)
 {
+#ifndef _SQUID_MSWIN_
 #ifdef _SQUID_NEXT_
     union wait status;
 #else
@@ -435,6 +484,7 @@ sig_child(int sig)
     } while (pid > 0 || (pid < 0 && errno == EINTR));
     signal(sig, sig_child);
 #endif
+#endif /* _SQUID_MSWIN_ */
 }
 
 const char *
@@ -443,24 +493,32 @@ getMyHostname(void)
     LOCAL_ARRAY(char, host, SQUIDHOSTNAMELEN + 1);
     static int present = 0;
     const struct hostent *h = NULL;
+    struct in_addr sa;
     if (Config.visibleHostname != NULL)
 	return Config.visibleHostname;
     if (present)
 	return host;
     host[0] = '\0';
-    if (Config.Sockaddr.http->s.sin_addr.s_addr != any_addr.s_addr) {
-	/*
-	 * If the first http_port address has a specific address, try a
-	 * reverse DNS lookup on it.
-	 */
-	h = gethostbyaddr((char *) &Config.Sockaddr.http->s.sin_addr,
-	    sizeof(Config.Sockaddr.http->s.sin_addr), AF_INET);
+    memcpy(&sa, &any_addr, sizeof(sa));
+    if (Config.Sockaddr.http && sa.s_addr == any_addr.s_addr)
+	memcpy(&sa, &Config.Sockaddr.http->s.sin_addr, sizeof(sa));
+#if USE_SSL
+    if (Config.Sockaddr.https && sa.s_addr == any_addr.s_addr)
+	memcpy(&sa, &Config.Sockaddr.https->http.s.sin_addr, sizeof(sa));
+#endif
+    /*
+     * If the first http_port address has a specific address, try a
+     * reverse DNS lookup on it.
+     */
+    if (sa.s_addr != any_addr.s_addr) {
+	h = gethostbyaddr((char *) &sa,
+	    sizeof(sa), AF_INET);
 	if (h != NULL) {
 	    /* DNS lookup successful */
 	    /* use the official name from DNS lookup */
 	    xstrncpy(host, h->h_name, SQUIDHOSTNAMELEN);
 	    debug(50, 4) ("getMyHostname: resolved %s to '%s'\n",
-		inet_ntoa(Config.Sockaddr.http->s.sin_addr),
+		inet_ntoa(sa),
 		host);
 	    present = 1;
 	    if (strchr(host, '.'))
@@ -468,7 +526,7 @@ getMyHostname(void)
 
 	}
 	debug(50, 1) ("WARNING: failed to resolve %s to a fully qualified hostname\n",
-	    inet_ntoa(Config.Sockaddr.http->s.sin_addr));
+	    inet_ntoa(sa));
     }
     /*
      * Get the host name and store it in host to return
@@ -487,7 +545,10 @@ getMyHostname(void)
 	if (strchr(host, '.'))
 	    return host;
     }
-    fatal("Could not determine fully qualified hostname.  Please set 'visible_hostname'\n");
+    if (opt_send_signal == -1)
+	fatal("Could not determine fully qualified hostname.  Please set 'visible_hostname'\n");
+    else
+	return ("localhost");
     return NULL;		/* keep compiler happy */
 }
 
@@ -505,6 +566,33 @@ safeunlink(const char *s, int quiet)
 	debug(50, 1) ("safeunlink: Couldn't delete %s: %s\n", s, xstrerror());
 }
 
+/* Should get called after any operation which may make the OS disable core dumps */
+void
+enableCoredumps(void)
+{
+    if (!Config.coredump_dir)
+	return;
+    if (strcmp(Config.coredump_dir, "none") == 0)
+	return;
+
+#if HAVE_PRCTL && defined(PR_SET_DUMPABLE)
+    /* Set Linux DUMPABLE flag */
+    if (prctl(PR_SET_DUMPABLE, 1, 0, 0, 0) != 0)
+	debug(50, 2) ("prctl: %s\n", xstrerror());
+#endif
+#if HAVE_SETRLIMIT && defined(RLIMIT_CORE)
+    /* Make sure coredumps are not limited */
+    {
+	struct rlimit rlim;
+
+	if (getrlimit(RLIMIT_CORE, &rlim) == 0) {
+	    rlim.rlim_cur = rlim.rlim_max;
+	    setrlimit(RLIMIT_CORE, &rlim);
+	}
+    }
+#endif
+}
+
 /* leave a privilegied section. (Give up any privilegies)
  * Routines that need privilegies can rap themselves in enter_suid()
  * and leave_suid()
@@ -514,18 +602,29 @@ void
 leave_suid(void)
 {
     debug(21, 3) ("leave_suid: PID %d called\n", (int) getpid());
+    if (Config.effectiveGroup) {
+#if HAVE_SETGROUPS
+	setgroups(1, &Config2.effectiveGroupID);
+#endif
+	if (setgid(Config2.effectiveGroupID) < 0)
+	    debug(50, 0) ("ALERT: setgid: %s\n", xstrerror());
+    }
     if (geteuid() != 0)
 	return;
     /* Started as a root, check suid option */
     if (Config.effectiveUser == NULL)
 	return;
-#if HAVE_SETGROUPS
-    setgroups(1, &Config2.effectiveGroupID);
-#endif
-    if (setgid(Config2.effectiveGroupID) < 0)
-	debug(50, 0) ("ALERT: setgid: %s\n", xstrerror());
     debug(21, 3) ("leave_suid: PID %d giving up root, becoming '%s'\n",
 	(int) getpid(), Config.effectiveUser);
+    if (!Config.effectiveGroup) {
+	if (setgid(Config2.effectiveGroupID) < 0)
+	    debug(50, 0) ("ALERT: setgid: %s\n", xstrerror());
+	if (initgroups(Config.effectiveUser, Config2.effectiveGroupID) < 0) {
+	    debug(50, 0) ("ALERT: initgroups: unable to set groups for User %s "
+		"and Group %u", Config.effectiveUser,
+		(unsigned) Config2.effectiveGroupID);
+	}
+    }
 #if HAVE_SETRESUID
     if (setresuid(Config2.effectiveUserID, Config2.effectiveUserID, 0) < 0)
 	debug(50, 0) ("ALERT: setresuid: %s\n", xstrerror());
@@ -536,6 +635,25 @@ leave_suid(void)
     if (setuid(Config2.effectiveUserID) < 0)
 	debug(50, 0) ("ALERT: setuid: %s\n", xstrerror());
 #endif
+#if LINUX_TPROXY
+    if (need_linux_tproxy) {
+	cap_user_header_t head = (cap_user_header_t) xcalloc(1, sizeof(cap_user_header_t));
+	cap_user_data_t cap = (cap_user_data_t) xcalloc(1, sizeof(cap_user_data_t));
+
+	head->version = _LINUX_CAPABILITY_VERSION;
+	head->pid = 0;
+	cap->inheritable = cap->permitted = cap->effective = (1 << CAP_NET_ADMIN) + (1 << CAP_NET_BIND_SERVICE) + (1 << CAP_NET_BROADCAST);
+	if (capset(head, cap) != 0) {
+	    xfree(head);
+	    xfree(cap);
+	    fatal("Error giving up capabilities");
+	}
+	xfree(head);
+	xfree(cap);
+    }
+#endif
+    /* Changing user ID usually blocks core dumps. Get them back! */
+    enableCoredumps();
 }
 
 /* Enter a privilegied section */
@@ -548,6 +666,7 @@ enter_suid(void)
 #else
     setuid(0);
 #endif
+    enableCoredumps();
 }
 
 /* Give up the posibility to gain privilegies.
@@ -559,7 +678,7 @@ no_suid(void)
     uid_t uid;
     leave_suid();
     uid = geteuid();
-    debug(21, 3) ("leave_suid: PID %d giving up root priveleges forever\n", (int) getpid());
+    debug(21, 3) ("no_suid: PID %d giving up root priveleges forever\n", (int) getpid());
 #if HAVE_SETRESUID
     if (setresuid(uid, uid, uid) < 0)
 	debug(50, 1) ("no_suid: setresuid: %s\n", xstrerror());
@@ -568,6 +687,7 @@ no_suid(void)
     if (setuid(uid) < 0)
 	debug(50, 1) ("no_suid: setuid: %s\n", xstrerror());
 #endif
+    enableCoredumps();
 }
 
 void
@@ -602,6 +722,7 @@ readPidFile(void)
 {
     FILE *pid_fp = NULL;
     const char *f = Config.pidFilename;
+    char *chroot_f = NULL;
     pid_t pid = -1;
     int i;
 
@@ -609,6 +730,12 @@ readPidFile(void)
 	fprintf(stderr, "%s: ERROR: No pid file name defined\n", appname);
 	exit(1);
     }
+    if (Config.chroot_dir && geteuid() == 0) {
+	int len = strlen(Config.chroot_dir) + 1 + strlen(f) + 1;
+	chroot_f = xmalloc(len);
+	snprintf(chroot_f, len, "%s/%s", Config.chroot_dir, f);
+	f = chroot_f;
+    }
     pid_fp = fopen(f, "r");
     if (pid_fp != NULL) {
 	pid = 0;
@@ -622,6 +749,7 @@ readPidFile(void)
 	    exit(1);
 	}
     }
+    safe_free(chroot_f);
     return pid;
 }
 
@@ -633,6 +761,7 @@ setMaxFD(void)
     /* try to use as many file descriptors as possible */
     /* System V uses RLIMIT_NOFILE and BSD uses RLIMIT_OFILE */
     struct rlimit rl;
+#if !defined(_SQUID_CYGWIN_)
 #if defined(RLIMIT_NOFILE)
     if (getrlimit(RLIMIT_NOFILE, &rl) < 0) {
 	debug(50, 0) ("setrlimit: RLIMIT_NOFILE: %s\n", xstrerror());
@@ -660,6 +789,7 @@ setMaxFD(void)
 	}
     }
 #endif
+#endif
 #else /* HAVE_SETRLIMIT */
     debug(21, 1) ("setMaxFD: Cannot increase: setrlimit() not supported on this system\n");
 #endif /* HAVE_SETRLIMIT */
@@ -826,7 +956,7 @@ dlinkDelete(dlink_node * m, dlink_list * list)
 }
 
 void
-kb_incr(kb_t * k, size_t v)
+kb_incr(kb_t * k, squid_off_t v)
 {
     k->bytes += v;
     k->kb += (k->bytes >> 10);
@@ -924,6 +1054,9 @@ int
 xrename(const char *from, const char *to)
 {
     debug(21, 2) ("xrename: renaming %s to %s\n", from, to);
+#if defined(_SQUID_OS2_) || defined(_SQUID_WIN32_)
+    remove(to);
+#endif
     if (0 == rename(from, to))
 	return 0;
     debug(21, errno == ENOENT ? 2 : 1) ("xrename: Cannot rename %s to %s: %s\n",
@@ -977,7 +1110,7 @@ parseEtcHosts(void)
 	    Config.etcHostsPath, xstrerror());
 	return;
     }
-#if defined(_SQUID_MSWIN_) || defined(_SQUID_CYGWIN_)
+#ifdef _SQUID_WIN32_
     setmode(fileno(fp), O_TEXT);
 #endif
     while (fgets(buf, 1024, fp)) {	/* for each line */
@@ -985,6 +1118,7 @@ parseEtcHosts(void)
 	char *addr;
 	if (buf[0] == '#')	/* MS-windows likes to add comments */
 	    continue;
+	strtok(buf, "#");	/* chop everything following a comment marker */
 	lt = buf;
 	addr = buf;
 	debug(1, 5) ("etc_hosts: line is '%s'\n", buf);
@@ -1048,7 +1182,18 @@ strwordtok(char *buf, char **t)
 	switch (ch) {
 	case '\\':
 	    p++;
-	    *d++ = ch = *p;
+	    switch (*p) {
+	    case 'n':
+		ch = '\n';
+		break;
+	    case 'r':
+		ch = '\r';
+		break;
+	    default:
+		ch = *p;
+		break;
+	    }
+	    *d++ = ch;
 	    if (ch)
 		p++;
 	    break;
@@ -1087,12 +1232,79 @@ strwordquote(MemBuf * mb, const char *str)
 	int l = strcspn(str, "\"\\");
 	memBufAppend(mb, str, l);
 	str += l;
-	while (*str == '"' || *str == '\\') {
+	switch (*str) {
+	case '\n':
+	    memBufAppend(mb, "\\n", 2);
+	    str++;
+	    break;
+	case '\r':
+	    memBufAppend(mb, "\\r", 2);
+	    str++;
+	    break;
+	case '\0':
+	    break;
+	default:
 	    memBufAppend(mb, "\\", 1);
 	    memBufAppend(mb, str, 1);
 	    str++;
+	    break;
 	}
     }
     if (quoted)
 	memBufAppend(mb, "\"", 1);
 }
+
+int
+getMyPort(void)
+{
+    if (Config.Sockaddr.http)
+	return ntohs(Config.Sockaddr.http->s.sin_port);
+#if USE_SSL
+    if (Config.Sockaddr.https)
+	return ntohs(Config.Sockaddr.https->http.s.sin_port);
+#endif
+    fatal("No port defined");
+    return 0;			/* NOT REACHED */
+}
+
+/*
+ * Set the umask to at least the given mask. This is in addition
+ * to the umask set at startup
+ */
+void
+setUmask(mode_t mask)
+{
+    static mode_t orig_umask = ~0;
+    if (orig_umask == (mode_t) ~ 0) {
+	/* Unfortunately, there is no way to get the current
+	 * umask value without setting it.
+	 */
+	orig_umask = umask(mask);
+    }
+    umask(mask | orig_umask);
+}
+
+/*
+ * xusleep, as usleep but accepts longer pauses
+ */
+int
+xusleep(unsigned int usec)
+{
+    /* XXX emulation of usleep() */
+    struct timeval sl;
+    sl.tv_sec = usec / 1000000;
+    sl.tv_usec = usec % 1000000;
+    return select(0, NULL, NULL, NULL, &sl);
+}
+
+void
+keepCapabilities(void)
+{
+#if LINUX_TPROXY
+    if (need_linux_tproxy) {
+	if (prctl(PR_SET_KEEPCAPS, 1, 0, 0, 0)) {
+	    debug(1, 1) ("Error - tproxy support requires capability setting which has failed.  Continuing without tproxy support\n");
+	}
+    }
+#endif
+}
